@@ -1,5 +1,7 @@
+use crate::error::FangError;
 use crate::queue::Queue;
 use crate::queue::Task;
+use crate::worker_pool::{SharedState, WorkerState};
 use diesel::pg::PgConnection;
 use diesel::r2d2::{ConnectionManager, PooledConnection};
 use log::error;
@@ -11,6 +13,7 @@ pub struct Executor {
     pub task_type: Option<String>,
     pub sleep_params: SleepParams,
     pub retention_mode: RetentionMode,
+    shared_state: Option<SharedState>,
 }
 
 #[derive(Clone)]
@@ -74,9 +77,14 @@ impl Executor {
             sleep_params: SleepParams::default(),
             retention_mode: RetentionMode::RemoveFinished,
             task_type: None,
+            shared_state: None,
         }
     }
 
+    pub fn set_shared_state(&mut self, shared_state: SharedState) {
+        self.shared_state = Some(shared_state);
+    }
+
     pub fn set_task_type(&mut self, task_type: String) {
         self.task_type = Some(task_type);
     }
@@ -91,12 +99,18 @@ impl Executor {
 
     pub fn run(&self, task: Task) {
         let result = self.execute_task(task);
-
         self.finalize_task(result)
     }
 
-    pub fn run_tasks(&mut self) {
+    pub fn run_tasks(&mut self) -> Result<(), FangError> {
         loop {
+            if let Some(ref shared_state) = self.shared_state {
+                let shared_state = shared_state.read()?;
+                if let WorkerState::Shutdown = *shared_state {
+                    return Ok(());
+                }
+            }
+
             match Queue::fetch_and_touch_query(&self.pooled_connection, &self.task_type.clone()) {
                 Ok(Some(task)) => {
                     self.maybe_reset_sleep_period();
@@ -301,7 +315,7 @@ mod executor_tests {
             executor.set_retention_mode(RetentionMode::KeepAll);
             executor.set_task_type("type1".to_string());
 
-            executor.run_tasks();
+            executor.run_tasks().unwrap();
         });
 
         std::thread::sleep(std::time::Duration::from_millis(1000));
