@@ -3,12 +3,14 @@
 #[macro_use]
 extern crate diesel;
 
+pub mod error;
 pub mod executor;
 pub mod queue;
 pub mod scheduler;
 pub mod schema;
 pub mod worker_pool;
 
+pub use error::FangError;
 pub use executor::*;
 pub use queue::*;
 pub use scheduler::*;
